from rougail.annotator.variable import Walk
from rougail.error import DictConsistencyError
from risotto.utils import _, multi_function
from warnings import warn


def _parse_kwargs(provider, dns, kwargs, index=None):
    if not isinstance(dns, list):
        raise Exception('pfff')
    values = {}
    for key, value in kwargs.items():
        if '_' not in key:
            raise Exception(f'unknown attribute {key} in calc_providers_global with provider {provider}')
        k, idx = key.rsplit('_', 1)
        values.setdefault(idx, {})[k] = value
    for idx, data in values.items():
        if index is not None and int(idx) != index:
            continue
        if 'dns' not in data:
            continue
        if isinstance(data['dns'], list):
            for ddns in data['dns']:
                if ddns in dns:
                    break
            else:
                continue
        elif data['dns'] not in dns:
            continue
#        del data['dns']
        yield data


@multi_function
def calc_providers_global(provider, multi, unique, value, suffix=None):
    if suffix is not None:
        return value[int(suffix)]
    return value


@multi_function
def calc_providers_follower(provider, multi, unique, dns, leader, index, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if 'leader' in data:
            if isinstance(data['leader'], list):
                for idx, leader_iter in enumerate(data['leader']):
                    if leader_iter == leader:
                        ret.append(data['value'][idx])
            elif data['leader']== leader:
                ret.extend(data['value'])
        else:
            if isinstance(data['value'], list):
                for v in data['value']:
                    if v not in ret:
                        ret.append(v)
            elif data['value'] not in ret:
                ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers_dynamic_follower(provider, multi, unique, dns, leader, index, suffix, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if data['dynamic'] != suffix:
            continue
        if 'leader' in data:
            for idx, leader_iter in enumerate(data['leader']):
                if leader_iter == leader:
                    if isinstance(data['value'], list):
                        ret.append(data['value'][idx])
                    else:
                        ret.append(data['value'])
        else:
            if isinstance(data['value'], list):
                for v in data['value']:
                    if v not in ret:
                        ret.append(v)
            elif data['value'] not in ret:
                ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers_dynamic(provider, multi, unique, dns, suffix, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if data['dynamic'] != suffix:
            continue
        if isinstance(data['value'], list):
            for v in data['value']:
                if not unique or v not in ret:
                    ret.append(v)
        elif data['value'] not in ret:
            ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers(provider, multi, unique, dns, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if isinstance(data['value'], list):
            commun_dns = list(set(data['dns']) & set(dns))
            if len(commun_dns) == 1:
                ret.append(data['value'][data['dns'].index(commun_dns[0])])
                continue
            for v in data['value']:
                if v in ret:
                    continue
                ret.append(v)

        elif data['value'] not in ret:
            ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


class Annotator(Walk):
    level = 9
    def __init__(self,
                 objectspace: 'RougailObjSpace',
                 *args):
        self.objectspace = objectspace
        self.get_suppliers_providers()
        self.dispatch_provider_supplier_to_zones()
        self.dispatch_provider_to_zones()
        self.convert_providers()
        self.convert_suppliers()

    def get_suppliers_providers(self) -> None:
        """ get supplier informations
        return something like:
        {'Host': ['host1.example.net', 'host2.example.net']}
        """
        self.suppliers = {}
        self.providers = {}
        for variable in self.get_variables():
            if not hasattr(variable, 'supplier') and not hasattr(variable, 'provider'):
                continue
            nf_dns = variable.path.split('.', 1)[0]
            server_name = self.objectspace.space.variables[nf_dns].doc
            # supplier
            if hasattr(variable, 'supplier') and ':' not in variable.supplier:
                server_names = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:server_names']
                zones = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:zones_name']
                s_data = {'option': variable,
                          'dns': server_name,
                          'path_prefix': nf_dns,
                          'server_names': server_names,
                          'zones': zones,
                          'providers_zone': {},
                          }
                if variable.supplier != 'Host' and not variable.supplier.startswith('Host:') and not variable.supplier.startswith('global:'):
                    if 'global:provider_zone' in self.objectspace.rougailconfig['risotto_globals'][server_name]:
                        s_data['provider_zone'] = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:provider_zone']
                self.suppliers.setdefault(variable.supplier, []).append(s_data)
                if not hasattr(variable, 'information'):
                    variable.information = self.objectspace.information(variable.xmlfiles)
                variable.information.supplier = variable.supplier
            # provider
            if hasattr(variable, 'provider'):
                provider_name = variable.provider
                p_data = {'option': variable,
                          'dns': server_name,
                          'provider_name': provider_name,
                          'path_prefix': nf_dns,
                          'suppliers_zone': {},
                          }
                if variable.provider != 'Host' and not variable.provider.startswith('Host:') and not variable.provider.startswith('global:'):
                    if 'global:provider_zone' in self.objectspace.rougailconfig['risotto_globals'][server_name]:
                        p_data['provider_zone'] = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:provider_zone']
                if self.objectspace.rougailconfig['risotto_globals'][server_name]['global:module_name'] == 'host':
                    server_names = [server_name]
                else:
                    server_names = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:server_names']
                p_data['server_names'] = server_names
                if self.objectspace.rougailconfig['risotto_globals'][server_name]['global:module_name'] != 'host':
                    p_data['zones'] = self.objectspace.rougailconfig['risotto_globals'][server_name]['global:zones_name']
                if ':' in provider_name:
                    p_data['is_main_provider'] = False
                    provider = provider_name.rsplit(':', 1)[0]
                else:
                    p_data['is_main_provider'] = True
                    provider = variable.provider
                self.providers.setdefault(provider, []).append(p_data)

    def dispatch_provider_supplier_to_zones(self):
        """calculate zone where provider and supplier communicate
        """
        self.providers_zone = {}
        for provider_name, p_datas in self.providers.items():
            for p_data in p_datas:
                if provider_name in ['global', 'Host'] or provider_name not in self.suppliers:
                    continue
                if not 'provider_zone' in p_data:
                    provider_zone = None
                else:
                    provider_zone = p_data['provider_zone']
                for s_data in self.suppliers[provider_name]:
                    if not provider_zone:
                        if provider_name.endswith('Client'):
                            p_server = provider_name[0:-6]
                            if p_server not in self.providers:
                                continue
                            for p_data_t in self.providers[p_server]:
                                if p_data_t['dns'] == s_data['dns'] and 'provider_zone' in p_data_t:
                                    zone = p_data_t['provider_zone']
                                    break
                            else:
                                continue
                        else:
                            continue
                    else:
                        zone = provider_zone
                    if zone not in s_data['zones']:
                        continue
                    s_data['providers_zone'][provider_name] = zone
                    p_data['suppliers_zone'].setdefault(provider_name, {})[s_data['dns']] = zone
                    self.providers_zone.setdefault(zone, set()).add(provider_name)

    def dispatch_provider_to_zones(self):
        """ add information with provider zone domain name
        """
        self.providers_zone = {}
        for provider_name, p_datas in self.providers.items():
            for p_data in p_datas:
                if provider_name in ['global', 'Host'] or provider_name not in self.suppliers:
                    continue
                if not 'provider_zone' in p_data:
                    continue
                provider_zone = p_data['provider_zone']
                family = self.objectspace.paths.get_variable(f"providers",
                                                             namespace=self.objectspace.rougailconfig['variable_namespace'],
                                                             force_path_prefix=p_data['path_prefix'],
                                                             )
                if not hasattr(family, 'information'):
                    family.information = self.objectspace.information(family.xmlfiles)
                name_in_zone = p_data['server_names'][p_data['zones'].index(provider_zone)]
                setattr(family.information, provider_name, name_in_zone)
                setattr(family.information, f'{provider_name}:zone', provider_zone)

    def convert_providers(self):
        for provider_name, providers_data in self.providers.items():
            for provider_data in providers_data:
                if provider_name != 'global' and provider_name not in self.suppliers:
                    continue
                # create a fill for this variable
                variable = provider_data['option']
                fill = self.objectspace.fill(variable.xmlfiles)
                new_target = self.objectspace.target(variable.xmlfiles)
                new_target.name = variable
                fill.target = [new_target]
                if provider_name == 'global':
                    fill.name = 'calc_providers_global'
                elif self.objectspace.paths.is_dynamic(variable):
                    if self.objectspace.paths.is_follower(variable):
                        fill.name = 'calc_providers_dynamic_follower'
                    else:
                        fill.name = 'calc_providers_dynamic'
                elif self.objectspace.paths.is_follower(variable):
                    fill.name = 'calc_providers_follower'
                else:
                    fill.name = 'calc_providers'
                fill.namespace = variable.namespace
                fill.index = 0
                # first parameter: the provider name (something link Host:incoming_ports)
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'provider'
                param.text = provider_data['provider_name']
                fill.param = [param]
                # second parameter: current variable is a multi variable?
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'multi'
                param.text = variable.multi
                param.type = 'boolean'
                fill.param.append(param)
                #
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'unique'
                param.text = variable.unique != "False"
                param.type = 'boolean'
                fill.param.append(param)
                #
                if self.objectspace.paths.is_follower(variable):
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = 'leader'
                    param.text = self.objectspace.paths.get_leader(variable)
                    param.propertyerror = False
                    param.type = 'variable'
                    fill.param.append(param)
                    try:
                        leader_provider = self.objectspace.paths.get_leader(variable).provider
                    except:
                        leader_provider = None
                    #
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = 'index'
                    param.type = 'index'
                    fill.param.append(param)
                if self.objectspace.paths.is_dynamic(variable):
                    # if dynamic: current suffix
                    # and add current DNS name, this is useful to known if supplier is link to this provider
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = 'suffix'
                    param.type = 'suffix'
                    fill.param.append(param)
                if provider_name != 'global':
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = 'dns'
                    param.text = provider_data['server_names']
                    fill.param.append(param)
                if provider_name == 'global':
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = 'value'
                    if provider_data['provider_name'] in self.objectspace.rougailconfig['risotto_globals'][provider_data['dns']]:
                        value = self.objectspace.rougailconfig['risotto_globals'][provider_data['dns']][provider_data['provider_name']]
                        param.text = value
                        if isinstance(value, bool):
                            param.type = 'boolean'
                    else:
                        param.text = provider_data['provider_name']
                        param.type = 'information'
                    fill.param.append(param)
                else:
                    # parse all supplier link to current provider
                    for idx, data in enumerate(self.suppliers[provider_name]):
                        if 'zones' in provider_data:
                            if provider_name not in data['providers_zone']:
                                continue
                            zone = data['providers_zone'][provider_name]
                            zidx = data['zones'].index(zone)
                            dns = data['server_names'][zidx]
                        else:
                            dns = data['dns']
                        option = data['option']
                        # if not provider, get the true option that we want have value
                        if not provider_data['is_main_provider']:
                            path_prefix = data['path_prefix']
                            try:
                                supplier_option = self.objectspace.paths.get_supplier(f'supplier:{provider_data["provider_name"]}', path_prefix)
                            except KeyError:
                                #warn(f'cannot find supplier "{provider_name}" for "{dns}"')
                                continue
                        # first of all, get the supplier name
                        param = self.objectspace.param(variable.xmlfiles)
                        param.name = f'dns_{idx}'
                        param.text = option
                        param.propertyerror = False
                        param.type = 'variable'
                        fill.param.append(param)
                        if not provider_data['is_main_provider'] and \
                                self.objectspace.paths.is_follower(variable):
                            param = self.objectspace.param(variable.xmlfiles)
                            param.name = f'leader_{idx}'
                            if fill.name == 'calc_providers_follower':
                                param.text = dns
                            else:
                                if self.objectspace.paths.is_follower(supplier_option):
                                    param.text = self.objectspace.paths.get_leader(supplier_option)
                                else:
                                    param.text = self.objectspace.paths.get_supplier(f'supplier:{leader_provider}', path_prefix)
                                param.propertyerror = False
                                param.type = 'variable'
                            fill.param.append(param)
                        # get the current DNS name for dynamic variable
                        if self.objectspace.paths.is_dynamic(variable):
                            param = self.objectspace.param(variable.xmlfiles)
                            param.name = f'dynamic_{idx}'
                            param.text = dns
                            fill.param.append(param)
                        # get the current value!
                        param = self.objectspace.param(variable.xmlfiles)
                        param.name = f'value_{idx}'
                        if provider_data['is_main_provider']:
                            param.text = dns
                        else:
                            param.text = supplier_option
                            param.propertyerror = False
                            param.type = 'variable'
                        fill.param.append(param)
                if not hasattr(self.objectspace.space.variables[provider_data['path_prefix']], 'constraints'):
                    self.objectspace.space.variables[provider_data['path_prefix']].constraints = self.objectspace.constraints(None)
                if not hasattr(self.objectspace.space.variables[provider_data['path_prefix']].constraints, 'fill'):
                    self.objectspace.space.variables[provider_data['path_prefix']].constraints.fill = []
                self.objectspace.space.variables[provider_data['path_prefix']].constraints.fill.append(fill)

    def convert_suppliers(self):
        for supplier_name, s_datas in self.suppliers.items():
            if supplier_name == 'Host':
                continue
            for s_data in s_datas:
                if supplier_name not in self.providers:
                    continue
                for p_data in self.providers[supplier_name]:
                    if s_data['dns'] == p_data['dns']:
                        # supplier and provider are in same machine
                        continue
                    if supplier_name not in p_data['suppliers_zone'] or s_data['dns'] not in p_data['suppliers_zone'][supplier_name]:
                        continue
                    # get the DNS name in supplier zone
                    zone = p_data['suppliers_zone'][supplier_name][s_data['dns']]
                    if zone not in p_data['zones']:
                        continue
                    zidx = p_data['zones'].index(zone)
                    dns = p_data['server_names'][zidx]
                    new_value = self.objectspace.value(None)
                    new_value.name = dns
                    s_data['option'].value = [new_value]
                    break
